"""Utility functions to retrieve lat-lon specific informations"""
import os
import os.path as osp
import json
from itertools import starmap, product
import numpy as np
import pandas as pd


__version__ = '0.0.3'

__author__ = 'Philipp S. Sommer'


#: Variables in the WorldClim v2.0 dataset
worldclim_variables = {
    'tmin': 'minimum temperature (degC)',
    'tmax': 'maximum temperature (degC)',
    'tavg': 'maximum temperature (degC)',
    'prec': 'precipitation (mm)',
    'srad': 'solar radiation (kJ m-2 day-1)',
    'wind': 'wind speed (m s-1)',
    'vapr': 'water vapor pressure (kPa)',
    }

#: The available resolutions for the WorldClim data
worldclim_resolutions = ['10m', '5m', '2.5m', '30s']


def get_data_dir():
    """Get the data directory where the downloads are stored.

    You can either set this with the ``LATLONDATA`` environment variable or we
    use ``$HOME/.local/share/latlon_utils``.

    Returns
    -------
    str
        The directory for the data. If it is not yet existing, it will be
        created

    See Also
    --------
    get_data_file: To retrieve the path to a file in this directory"""
    ret = os.getenv('LATLONDATA', osp.join(osp.expanduser("~"), '.local',
                                           'share', 'latlon_utils'))
    if not osp.exists(ret):
        os.makedirs(ret)
    return ret


def get_wc_resolution(res):
    """Get the resolution used for the WorldClim data

    Parameters
    ----------
    res: {'10m', '5m', '2.5m', '30s'}
        If None, the ``LATLONRES`` environment variable is used. If this is not
        set, we set the default resolution to '10m'"""
    return res or os.getenv("LATLONRES", '10m')


def get_data_file(fname, download=True):
    """Get the path of a data file in the data directory

    Parameters
    ----------
    fname: str
        The file name in the data directory (see :func:`get_data_dir`)
    download: bool
        If True and the file is missing, it will be downloaded"""
    ret = osp.join(get_data_dir(), fname)
    if not osp.exists(ret):
        if fname == 'countries.geojson':
            from latlon_utils.download import download_geo_countries
            download_geo_countries()
            return ret
        elif fname in starmap(
                '{}_{}.nc'.format,
                product(worldclim_variables, worldclim_resolutions)):
            from latlon_utils.download import download_wc_variable
            var, res = osp.splitext(fname)[0].split('_')
            download_wc_variable(var, res=res)
            return ret
        else:
            raise ValueError(
                "Could not find %s in %s!" % (fname, get_data_dir()))
    return ret


def get_country(lat, lon):
    """Get the country at the specific latitude and longitude

    Parameters
    ----------
    lat: float or np.ndarray of floats
        The latitude between -90 and 90
    lon: float or np.ndarray of floats
        The longitude between -180 and 360

    Returns
    -------
    str or list of str
        The Country name(s) or ``'unknown'`` if no country could be found at
        the given lat-lon combination

    Examples
    --------
    Get the country for 50 degrees north and 10 degrees east::

        >>> from latlon_utils import get_country

        >>> get_country(50, 10)
        'Germany'
    """
    from shapely.geometry import shape, Point
    from shapely.prepared import prep

    assert np.shape(lat) == np.shape(lon)

    with open(get_data_file('countries.geojson')) as f:
        data = json.load(f)

    countries = {}
    for feature in data["features"]:
        geom = feature["geometry"]
        country = feature["properties"]["ADMIN"]
        countries[country] = prep(shape(geom))

    def get_country(lat, lon):
        point = Point(lon, lat)
        for country, geom in countries.items():
            if geom.contains(point):
                return country

        return "unknown"

    if not np.ndim(lat):
        return get_country(lat, lon)
    else:
        return list(starmap(get_country, zip(lat, lon)))


def test_get_country():
    assert get_country(50, 10) == 'Germany'


def get_climate(lat, lon, variables=['tavg', 'prec'], res=None):
    """Get the country at the specific latitude and longitude

    Parameters
    ----------
    lat: float or np.ndarray of floats
        The latitude between -90 and 90
    lon: float or np.ndarray of floats
        The longitude between -180 and 360
    variables: list of str
        The variables to include. By default average temperature (tavg) and
        precipitation (prec). Avaiable datasets in the WorldClim v2.0 dataset
        are

        tmin
            minimum temperature (°C)
        tmax
            maximum temperature (°C)
        tavg
            maximum temperature (°C)
        prec
            precipitation (mm)
        srad
            solar radiation (kJ m-2 day-1)
        wind
            wind speed (m s-1)
        vapr
            water vapor pressure (kPa)
    res: str
        The resolution to use. If None, it defaults to the
        ``LATLONRES`` environment variable or ``'10m'``

    Returns
    -------
    pandas.DataFrame or pandas.Series
        Climate information for the given cells. The index is a
        :class:`pandas.MultiIndex` with the first level being `lat` and the
        second `lon`, the columns are a :class:`pandas.MultiIndex` with the
        first variable being one of ``'tavg'`` and ``'prec'`` and the second
        are the monthly, seasonal and annual means. To access, for example, the
        temperature in july, you can write ``df['tavg', 'jul']``

        If `lat` is not an array, a pandas.Series is returned.

    Examples
    --------
    Get the climate for 50 degrees north and 10 degrees east::

        >>> from latlon_utils import get_climate

        # limit the number of columns printed by pandas
        >>> import pandas; pandas.options.display.max_columns = 5

        >>> get_climate(50, 10)
        tavg  jan     0.044739
              feb     0.974976
              mar     4.705505
              apr     8.232239
              mai    13.150024
              jun    16.012268
              jul    17.958984
              aug    17.828735
              sep    13.779480
              oct     8.787476
              nov     4.039001
              dec     1.430237
              djf     0.816650
              mam     8.695923
              jja    17.266663
              son     8.868652
              ann     8.911972
        prec  jan    48.000000
              feb    42.000000
              mar    44.000000
              apr    44.000000
              mai    56.000000
              jun    68.000000
              jul    65.000000
              aug    52.000000
              sep    47.000000
              oct    52.000000
              nov    52.000000
              dec    59.000000
              djf    49.666667
              mam    48.000000
              jja    61.666667
              son    50.333333
              ann    52.416667
        Name: (50, 10), dtype: float64


        >>> get_climate(50, 10)['tavg', 'djf']
        0.816650390625

        >>> get_climate([10, 11], [50, 51])
                    tavg             ...       prec
                     jan        feb  ...        son       ann
        lat lon                        ...
        10  50   21.810730  22.687988  ...  10.666667  6.833333
        11  51   24.617249  24.678040  ...   7.666667  3.750000
        <BLANKLINE>
        [2 rows x 34 columns]
    """
    import netCDF4 as nc
    assert np.shape(lat) == np.shape(lon)

    def find_closest(arr, vals):
        idx = np.searchsorted(arr, vals)
        diff = arr[idx] - vals
        diffprev = vals - arr[idx - 1]
        idx[idx == arr.size] = arr.size - 1
        idx[(idx > 0) & (idx < arr.size) & (diff > diffprev)] -= 1
        return idx

    squeeze = False
    if np.ndim(lat) == 0:
        lat = [lat]
        lon = [lon]
        squeeze = True

    lat = np.asarray(lat)
    lon = np.asarray(lon)
    lon = np.where(lon > 180, lon - 360, lon)

    months = ['jan', 'feb', 'mar', 'apr', 'mai', 'jun', 'jul', 'aug', 'sep',
              'oct', 'nov', 'dec']

    cols = months + ['djf', 'mam', 'jja', 'son', 'ann']

    ret = pd.DataFrame(
        index=pd.MultiIndex.from_tuples(list(zip(lat, lon)),
                                        names=['lat', 'lon']),
        columns=pd.MultiIndex.from_product([variables, cols]),
        dtype=float)

    res = get_wc_resolution(res)

    data_files = [get_data_file(v + '_' + res + '.nc') for v in variables]

    for v, fname in zip(variables, data_files):
        with nc.Dataset(fname) as nco:
            nco.set_auto_mask(False)
            idx_lon = find_closest(nco.variables['lon'][:], lon)
            idx_lat = nco.variables['lat'].shape[0] - 1 - find_closest(
                nco.variables['lat'][::-1], lat)
            for i, (j, k) in enumerate(zip(idx_lat, idx_lon)):
                ret.loc[slice(i, i+1), (v, months)] = nco.variables[v][:, j, k]

    # compute seasonal averages
    for var in variables:
        ret[var, 'djf'] = ret[var][['jan', 'feb', 'dec']].mean(axis=1)
        ret[var, 'mam'] = ret[var][['mar', 'apr', 'mai']].mean(axis=1)
        ret[var, 'jja'] = ret[var][['jun', 'jul', 'aug']].mean(axis=1)
        ret[var, 'son'] = ret[var][['sep', 'oct', 'nov']].mean(axis=1)
        ret[var, 'ann'] = ret[var][months].mean(axis=1)
    if squeeze:
        return ret.T.iloc[:, 0]
    return ret


def test_get_climate():
    """Test the :func:`get_climate` function"""
    import netCDF4 as nc
    from numpy.testing import assert_allclose
    lat, lon = 50, 10
    with nc.Dataset(get_data_file('tavg_10m.nc')) as nco:
        idx_lat = pd.Index(nco.variables['lat'][:]).get_loc(
            lat, method='nearest')
        idx_lon = pd.Index(nco.variables['lon'][:]).get_loc(
            lon, method='nearest')
        tmonth = nco.variables['tavg'][:, idx_lat, idx_lon]
    with nc.Dataset(get_data_file('prec_10m.nc')) as nco:
        pmonth = nco.variables['prec'][:, idx_lat, idx_lon]

    ref = np.r_[tmonth, tmonth[[0, 1, -1]].mean(), tmonth[2:5].mean(),
                tmonth[5:8].mean(), tmonth[8:11].mean(), tmonth.mean(),
                pmonth, pmonth[[0, 1, -1]].mean(), pmonth[2:5].mean(),
                pmonth[5:8].mean(), pmonth[8:11].mean(), pmonth.mean()]

    assert_allclose(get_climate(lat, lon), ref)



import argparse
import os
import os.path as osp
from urllib import request
import xarray as xr
import numpy as np
import zipfile
import glob
import re
import datetime as dt
import tempfile
from latlon_utils import (
    get_data_dir, worldclim_variables, worldclim_resolutions, __version__)


wc_base_url = 'http://biogeo.ucdavis.edu/data/worldclim/v2.0/tif/base/'

SILENT = False

unit_patt = re.compile(r'\(.+\)')


def download_wc_variable(name, outdir=None, res='5m', lat=None, lon=None):
    """Download a WorldClim variable and transform it to a netcdf file

    Parameters
    ----------
    name: str
        The variable name. One of

        tmin
            minimum temperature (°C)
        tmax
            maximum temperature (°C)
        tavg
            maximum temperature (°C)
        prec
            precipitation (mm)
        srad
            solar radiation (kJ m-2 day-1)
        wind
            wind speed (m s-1)
        vapr
            water vapor pressure (kPa)
    outdir: str
        The target directory. If None, the default data directory for the
        latlon_utils (see the :func:`latlon_utils.get_data_dir` function) is
        used
    res: str
        The resolution string. One out of ``'10m', '5m', '2.5m'`` and ``'30s'``
    lat: indexer
        An indexer to use for the latitudes to extract a subset of the data
        Can be anything valid for the :meth:`xarray.DataArray.sel` method.
    lon: indexer
        An indexer to use for the latitudes to extract a subset of the data.
        Can be anything valid for the :meth:`xarray.DataArray.sel` method.

    Returns
    -------
    str
        The path to the downloaded netCDF file
    """
    if outdir is None:
        outdir = get_data_dir()

    if not osp.exists(outdir):
        os.makedirs(outdir)
    base = 'wc2.0_%s_%s.zip' % (res, name)
    outfile = osp.join(outdir, name + '_' + res + '.nc')

    with tempfile.TemporaryDirectory(prefix='worldclim_') as download_dir:
        download_target = osp.join(download_dir, base)

        url = wc_base_url + base

        if not SILENT:
            print('Downloading ' + url)
        request.urlretrieve(url, download_target)

        if not SILENT:
            print('Extracting ' + download_target)
        with zipfile.ZipFile(download_target) as f:
            f.extractall(download_dir)

        tiffs = sorted(glob.glob(osp.join(download_dir,
                                          'wc2.0_%s_%s_??.tif' % (res, name))))
        da = xr.concat(list(map(xr.open_rasterio, tiffs)),
                       dim=xr.Variable(('month', ), np.arange(1, 13)))
        da.encoding = dict(zlib=True, complevel=4, least_significant_digit=4)
        da.name = name

        if not SILENT:
            print("Saving as netcdf file to " + outfile)
        sel = {'band': 1}
        if lat is not None:
            sel['y'] = lat
        if lon is not None:
            sel['x'] = lon
        da.attrs['long_name'] = unit_patt.sub(
            '', worldclim_variables[name]).strip()
        da.attrs['units'] = unit_patt.search(
            worldclim_variables[name]).group()[1:-1]
        da.attrs['history'] = (
            '%s: Downloaded %s data with latlon-utils (%s) from %s' % (
                dt.datetime.now().isoformat(), name, __version__, url))
        da.attrs['reference'] = (
            'Fick, S.E. and R.J. Hijmans, 2017. Worldclim 2: New 1-km spatial '
            'resolution climate surfaces for global land areas. International '
            'Journal of Climatology.')
        da.sel(**sel).rename({'x': 'lon', 'y': 'lat'}).to_netcdf(outfile)
    return outfile


def download_geo_countries(outdir=None):
    """Download the datasets/geo-countries dataset

    Parameters
    ----------
    outdir: str
        The target directory. If None, the default data directory for the
        latlon_utils (see the :func:`latlon_utils.get_data_dir` function) is
        used

    Returns
    -------
    str
        The path to the downloaded file"""
    if outdir is None:
        outdir = get_data_dir()

    if not osp.exists(outdir):
        os.makedirs(outdir)
    download_target = osp.join(outdir, 'countries.geojson')

    url = ("https://raw.githubusercontent.com/datasets/geo-countries/master/"
           "data/countries.geojson")

    if not SILENT:
        print('Downloading %s to %s' % (url, download_target))
    request.urlretrieve(url, download_target)
    return download_target


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'outdir',
        help=("The target directory where to download the data. "
              "Default: %(default)s"), nargs='?',
        default=get_data_dir())
    parser.add_argument(
        '-res', '--worldclim-resolution', default='10m',
        choices=worldclim_resolutions,
        help="The resolution for the WorldClim data. Default: %(default)s")
    parser.add_argument(
        '-v', '--worldclim-variables', nargs='+', default=['tavg', 'prec'],
        help="WorldClim variables to download. Default: %(default)s",
        choices=['all'] + list(worldclim_variables))
    parser.add_argument('-lat', nargs=2, type=float,
                        help='Minimum and maximum latitude')
    parser.add_argument('-lon', nargs=2, type=float,
                        help='Minimum and maximum longitude')
    parser.add_argument('-no-wc', '--no-worldclim', action='store_true',
                        help="Skip the download of WorldClim data")
    return parser


def main(args=None):
    parser = get_parser()
    args = parser.parse_args(args)

    # update global variables
    lon = args.lon and slice(*sorted(args.lon))
    lat = args.lat and slice(*sorted(args.lat)[::-1])
    outdir = args.outdir

    # download WorldClim data
    if not args.no_worldclim:
        if 'all' in args.worldclim_variables:
            variables = list(worldclim_variables)
        else:
            variables = args.worldclim_variables
        for v in variables:
            download_wc_variable(v, outdir, lat=lat, lon=lon,
                                 res=args.worldclim_resolution)

    # download countries.geojson
    download_geo_countries(outdir)


def test_command_line():
    with tempfile.TemporaryDirectory(prefix='worldclim_') as test_dir:
        main((test_dir + ' -v tavg -lat 20 30 -lon 20 30 -res 10m').split())
        assert osp.exists(osp.join(test_dir, 'tavg_10m.nc'))
        assert osp.exists(osp.join(test_dir, 'countries.geojson'))


if __name__ == '__main__':
    main()



import os.path as osp
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='latlon-utils',
      version='0.0.3',
      description=('Retrieve WorldClim climate and other information for '
                   'lat-lon grid cells'),
      long_description=readme(),
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: GIS',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: OS Independent',
      ],
      keywords='worldclim geo-countries latitude longitude',
      url='https://github.com/Chilipp/latlon-utils',
      author='Philipp Sommer',
      author_email='philipp.sommer@unil.ch',
      license="GPLv3",
      packages=find_packages(exclude=['docs', 'tests*', 'examples']),
      install_requires=[
          'netCDF4',
          'shapely',
          'pandas',
      ],
      package_data={'latlon-utils': [
          osp.join('latlon_utils', 'data', '*'),
          ]},
      include_package_data=True,
      tests_require=['pytest', 'rasterio', 'xarray', 'pytest-cov'],
      cmdclass={'test': PyTest},
      zip_safe=False)
